#include "predict.h"
#include "formulas.h"

/***************************************************************************************
  QPoint_Cdt_Input_File:  
  		Outputs the query point info into a file formatted so that TRIANGLE can
	read it.
***************************************************************************************/ 
void QPoint_Cdt_Input_File()
{
  int     		i;
  FILE     		*fpout;
  VERTEX_TYPE* 	pQP = (VERTEX_TYPE*) gpQP_Tri[0];
  
  fpout=fopen("qpoints.cdt.poly","w");
    if (fpout==NULL)    ERROR_Message("Error opening file: ", "qpoints.cdt.poly"); 
  fprintf(fpout,"%li %i %i %i\n", gaiNumQP[0],2,0,0);
  for(i=0; i < gaiNumQP[0]; i++)
  {
    if (gi3DOn)
      fprintf(fpout,"%i %f %f %f 0\n", i, pQP->fx, pQP->fy, pQP->fz);
    else
      fprintf(fpout,"%i %f %f 0 0\n", i, pQP->fx, pQP->fy);
    pQP++;
  }
  fprintf(fpout, "0 0\n");
  fprintf(fpout, "0\n");
  
  fclose(fpout);
}
int	Terrain_Point(POINT_TYPE sP)
{
	long	i, iYes=0;

	for(i=0; i<giNumTerr; i++)
		if (Macro_FloatEqual(sP.fx, (gpsTerrain+i)->fx) &&
			Macro_FloatEqual(sP.fy, (gpsTerrain+i)->fy) &&
			Macro_FloatEqual(sP.fz, (gpsTerrain+i)->fz))
			{ iYes = 1; i = giNumTerr; }

	return(iYes);
} 
/***************************************************************************************
	Wall_Edge_At_Level:
		Used by Obst_Cdt_Input_File to check if a wall has an edge on the z=0
	plane.  If so, V1 and V2 contain the edge point ID numbers and it returns 1.
	Otherwise, returns 0.
***************************************************************************************/ 
int Wall_Edge_At_Level(long* piVert, short iNumVert, long* piV1, long* piV2)
{
	short		j;
	long		k;
	POINT_TYPE 	sP1, sP2;

	*piV1 = -1; *piV2 = -1;

	if (giCL_Terrain) {
		for (k=0; k<giNumTerr; k++) {
		  if (k==*piV1) k = k+1;
			sP2 = *(gpsTerrain+k);
     
			for(j=0; j<iNumVert; j++) {
				sP1 = *(gpsObst_Vertices + *(piVert+j));
				if (Macro_FloatEqual(sP1.fx, sP2.fx) && Macro_FloatEqual(sP1.fy, sP2.fy) && Macro_FloatEqual(sP1.fz, sP2.fz))
				{
					if (*piV1==-1) { 
						*piV1 = k;
						k=0;
					}
					else {
						*piV2 = k;
						j = iNumVert; k = giNumTerr;
					}
				}
			}
		}
	}
	else {
		for(j=0; j<iNumVert; j++) {
			if ((gpsObst_Vertices + *(piVert+j))->fz < 30)
			{
				if (*piV1==-1)	
					*piV1 = *(piVert+j);
				else {
					*piV2 = *(piVert+j);
					j = iNumVert;
				}
			}
		}
	}
	if (*piV2 != -1)	
		return(1);
	else
		return(0);
}
	/* returns the position of the vertex in the given array */
long Find_Vertex_ID(VERTEX_TYPE iV1, VERTEX_TYPE asV[], long iNum)
{
	long i, iDone = -1;
  
  	for(i=0; i<iNum; i++)
    	if (Macro_PointsEqual3D(iV1, asV[i])) {
			iDone = i;
			i = iNum;
      	}
	if (iDone == -1) {printf("iDone!"); cc=getchar(); }
  	
	return(iDone);
}

/***************************************************************************************
  Obst_Cdt_Input_File:
  		Outputs the obstacle info into a file formatted so that TRIANGLE can
	read it.
***************************************************************************************/ 
void Obst_Cdt_Input_File()
{
  	FILE            *fpout;
  	long			i=0, iV1, iV2, iCount, iNum;
  	short			iNumVert;
  	long*			piVert;
  	VERTEX_TYPE*    pV;
  	long            aiE[MAX_EDGES][2];
      
  	fpout=fopen("obst.cdt.poly","w");
  	if (fpout==NULL)    ERROR_Message("Error opening file: ", "obst.cdt.poly"); 

	if (giCL_Terrain)
		iNum = giNumTerr;
	else {
		iNum=0; pV = gpsObst_Vertices;
		for(i=0; i<giNumObstVert; i++)
			if ((pV+i)->fz == 0) iNum++;
	}
 		/* set global variable */

  	gaiNum_Obst[0] = iNum;
  	gpObst_Tri[0] = (void*) calloc(iNum, sizeof(VERTEX_TYPE));
	
	if (giCL_Terrain) 
	  for(i=0; i<iNum; i++) { *(((VERTEX_TYPE*) gpObst_Tri[0]) + i) = *(gpsTerrain + i);}

 	pV = (VERTEX_TYPE*) gpObst_Tri[0];
  
 	fprintf(fpout,"%li %i %i %i\n", iNum,2,0,0);
	if (giCL_Terrain)	iNum = giNumTerr;	else iNum = giNumObstVert;
	iCount=0;

 	for(i=0; i<iNum; i++) {
		if (giCL_Terrain &&(gpsTerrain+i)->fz < 30 ) 
	  		fprintf(fpout,"%li %f %f\n", i, (gpsTerrain+i)->fx, (gpsTerrain+i)->fy);
		else if ((gpsObst_Vertices+i)->fz ==0) {
 			pV->fx = (gpsObst_Vertices+i)->fx; pV->fy = (gpsObst_Vertices+i)->fy; pV->fz = (float) 0;
	  		fprintf(fpout,"%li %f %f\n", iCount, pV->fx, pV->fy);
			pV++; iCount++;
		}
	}
  
		/* Count the number of valid wall edges that lie on the z=0 plane */  
		/* or the number that match the height field data if giCL_Terrain==1 */
 	iCount=0;
  	for(i=0; i<giNumObstWalls; i++)
	{  
		iNumVert = (gpsObst_Walls+i)->iNumVert;
		piVert = (gpsObst_Walls+i)->piVert;
		if (Wall_Edge_At_Level(piVert, iNumVert, &iV1, &iV2)) {
if (giCL_Terrain) {
  aiE[iCount][0] = iV1; aiE[iCount][1] = iV2;

}
else {
			aiE[iCount][0] = Find_Vertex_ID(*(gpsObst_Vertices + iV1), gpObst_Tri[0], gaiNum_Obst[0]);
		    aiE[iCount][1] = Find_Vertex_ID(*(gpsObst_Vertices + iV2), gpObst_Tri[0], gaiNum_Obst[0]); }
		    iCount++;
		 }
	} 
  	fprintf(fpout,"%li 0\n", iCount); 

	for(i=0; i<iCount; i++)
	  fprintf(fpout,"%li %li %li 3\n", i, aiE[i][0], aiE[i][1]);
		 
  	fprintf(fpout,"0\n"); /* no holes */
printf("iNum %i\n", iNum);
  	fclose(fpout);
}

/***************************************************************************************
	Find_Edge: 
		This is called to locate the edge that is associated with the triangle
	 iTri_Num.  If the edge is already in the edge data structure then this is the
	 edge's second incident triangle and it is then added as info to the edge data.
	 Otherwise, the edge has not been added yet so 0 is returned.

		Finds the edge ID # that is defined by the two points iP1, iP2
		Adds the tri ID# number to the edge's info

	returns the edge ID# or 0
***************************************************************************************/ 
long Find_Edge(EDGE_TYPE* psE, long iP1, long iP2, long iTri_Num, long iNum_Edges)
{
  long            i, iRet=-1;

  if(iP1>iP2)
  {
    i=iP2;
    iP2=iP1;
    iP1=i;
  }
  for(i=0;i<iNum_Edges;i++)
  {
    if((psE+i)->iPoint[0] == iP1)
    {
      if((psE+i)->iPoint[1] == iP2)
      {
        (psE+i)->iTri[1] = iTri_Num;
        iRet = i;
	i = iNum_Edges;
      }
    }
  }
  return(iRet);
}

/***************************************************************************************
	Add_Edge:
		Adds the edge info to the edge data structure.  The edge should be an
	edge that has not been added yet.
***************************************************************************************/ 
void Add_Edge(EDGE_TYPE* psEdge, long iP1, long iP2, long iTri_Num, long iNum_Edge)
{
  long i;

  if(iP1 > iP2)
  {
    i = iP2;
    iP2 = iP1;
    iP1 = i;
  }
  (psEdge+iNum_Edge)->iPoint[0] = iP1;
  (psEdge+iNum_Edge)->iPoint[1] = iP2;
  (psEdge+iNum_Edge)->iTri[0] = iTri_Num;
  (psEdge+iNum_Edge)->iTri[1] = -1;        /* will stay -1 if edge is on the boundary */
  (psEdge+iNum_Edge)->iWallType = 0;		   /* 0 means it is not a wall edge */
  (psEdge+iNum_Edge)->iThickness = 0;
}

/***************************************************************************************
	Check_Edge:
		Uses Find_Edge to determine whether the edge has already been added to the
	edge array or not.  If not, it adds it.  

	returns the Edge_ID in either case.
***************************************************************************************/ 
long Check_Edge(EDGE_TYPE* psE, long iP1, long iP2, long iNum_Tri, long* piNum_Edges)
{
  long iEdge_Id;
 
  iEdge_Id = Find_Edge(psE, iP1, iP2, iNum_Tri, *piNum_Edges);

  if (iEdge_Id == -1)
    {
      Add_Edge(psE, iP1, iP2, iNum_Tri, *piNum_Edges);     
      (*piNum_Edges)++;
      return(*piNum_Edges - 1);   
    }
  else
    {
      return(iEdge_Id);
    }
}

/***************************************************************************************
	Setup_Tri_Data_Structure:
		The output from the TRIANGLE program is a file that defines the triangles
	and there associated points, identified by ID #'s.  This routine reads the file
	and then sets up the three data structures for the triangle data structure,
	Edges and Triangles.  The Vertex array has already been created.
		pPET is a pointer to the array that should store the pointers to each data
	structure.  This function assigns the values to the array.
***************************************************************************************/ 
void Setup_Tri_Data_Structure(char* pFile_Name, void* apPET[], long* piNum_Tri, long* piNum_Edge) 
{
  FILE                *fp;
  int                 junk;
  long                i, iNum_Triangles, iNum_Edges;
  TRIANGLE_TYPE* psTri, *psTtemp;
  EDGE_TYPE      *psEdge;
 
     /*********** opens and reads file created by TRIANGLE ************/
  fp=fopen(pFile_Name, "r");
    if (fp==NULL)    ERROR_Message("Could not open file for reading: ", pFile_Name); 
  fscanf(fp,"%li %i %i\n",&iNum_Triangles,&junk,&junk);
  
            /***** allocate space for the storage of the triangle data structures *****/
  psTri = (TRIANGLE_TYPE*) calloc(iNum_Triangles, sizeof(TRIANGLE_TYPE));
  psTtemp = psTri;
  psEdge = (EDGE_TYPE*) calloc((iNum_Triangles * 2), sizeof(EDGE_TYPE));

  for(i=0;i<iNum_Triangles;i++)
  {
    fscanf(fp,"%li %li %li %li\n",&junk, &(psTtemp->iPoint[0]),&(psTtemp->iPoint[1]),&(psTtemp->iPoint[2]));
    psTtemp++;
  }
  fclose(fp);
     /*******************************************************/

  iNum_Edges=0;
  psTtemp = psTri;

  for(i=0;i<iNum_Triangles;i++)
  {
    psTtemp->iEdge[0] = Check_Edge(psEdge, psTtemp->iPoint[0], psTtemp->iPoint[1], i, &iNum_Edges);
    psTtemp->iEdge[1] = Check_Edge(psEdge, psTtemp->iPoint[1], psTtemp->iPoint[2], i, &iNum_Edges);
    psTtemp->iEdge[2] = Check_Edge(psEdge, psTtemp->iPoint[2], psTtemp->iPoint[0], i, &iNum_Edges);
    psTtemp++;
  }

  apPET[1] = (void*) psEdge;
  apPET[2] = (void*) psTri;
  *piNum_Tri = iNum_Triangles;
  *piNum_Edge = iNum_Edges;
}

/***************************************************************************************
  Compute_Trans_QPoints:
  		This function computes and stores with the transmitter the list of query point
	IDs that could possibly be heard by this transmitter.
		It uses the Free Space loss calculations for determining who can 'hear' the
	transmiter since it will provide the most optimistic prediction of all the methods.
***************************************************************************************/
void Compute_Trans_QPoints()
{
 TRANS_TYPE*   	psT = gpsTrans_Data;
 QP_TYPE* 		psQP;
 int          	t, i, iNum_Points;
 double       	fPower, fDist;
 char         	acPoints[MAX_TRANS_POINTS];
 long*        	pL;
 
 for(t=0; t<giNum_Trans; t++)
   {
     psQP = gpsQPoints;
     
     gfFree = Square(gfC / ((double) 4 * PIE * psT->fFreq_HRange));

     iNum_Points = 0;
     for(i=1; i <= giNumQP; i++)
     {
	 	fDist = Macro_Dist(psT->sLoc, psQP->sLoc);
     	fPower = Macro_Free_Space_Calc(psT->fMax_Power, fDist); 
		
	 	if (fPower > POWER_THRESHOLD_HEAR)
	   	{
	     	iNum_Points++;
	     	acPoints[i] = 'y';
	   	}
	 	else {
	   		acPoints[i]='n';
		}
      	psQP++;
	 }

     psT->iNum_Points = iNum_Points;
     psT->piQPoints = (long*) calloc(iNum_Points, sizeof(long));
     pL = psT->piQPoints;     
     for(i=0;i<giNumQP;i++)
     {
	 	if (acPoints[i+1]=='y')
	   	{
	    	*pL = i;
	    	pL++;
	  	}
     }
     psT++;
   }
}
/***************************************************************************************
	Compute_Terrain_Tri_Planes:
		When the terrain is not flat each triangle in the Obst_Tri triangulation
	will define a plane.  To speed reflection calculations these planes are 
	precomputed and stored in gpsTerrPlanes.
***************************************************************************************/
PLANE_TYPE* Compute_Terrain_Tri_Planes()
{
	long			i, j;
	VERTEX_TYPE*	pV = (VERTEX_TYPE*) gpObst_Tri[0];
	long			iNumV = gaiNum_Obst[0];
	TRIANGLE_TYPE*	pT = (TRIANGLE_TYPE*) gpObst_Tri[2];
	long			iNumT = gaiNum_Obst[2];
	PLANE_TYPE*		psPl = (PLANE_TYPE*) calloc(iNumT, sizeof(PLANE_TYPE));
	POINT_TYPE		sP[3], sPdir;

	sPdir.fx = (float) 0;  sPdir.fy = (float) 0;  sPdir.fz = (float) 1000;

	for(i=0; i<iNumT; i++) {
		for(j=0; j<3; j++)	
			sP[j] = *(pV + (pT+i)->iPoint[j]);
		*(psPl+i) = Plane_Point_Normal_Plus(sP[0], sP[1], sP[2], sPdir);
	}

	return(psPl);
}
/***************************************************************************************
	Edge_Real_or_Added:
		Checks the Obst_Tri edges to see which edges are walls or if they were added
	during the triangulation.  Currently O(n^2) but might be able to avoid this
	by using the attribute fields for edges in the CDT file.  TRIANGLE might
	keep track and thus the WallType and Thickness could be tracked through
	the triangulation process.
***************************************************************************************/
void Edge_Real_or_Added(void* gpObst_Tri[], long gaiNum_Obst[])
{
	long			i, j;
	short			k, iNumMatch;
	VERTEX_TYPE*	pV = (VERTEX_TYPE*) gpObst_Tri[0];
	long			iNumV = gaiNum_Obst[0];
	EDGE_TYPE*		pE = (EDGE_TYPE*) gpObst_Tri[1];
	long			iNumE = gaiNum_Obst[1];
	VERTEX_TYPE		sV_Tri, sV_Wall;
	WALL_TYPE*		pW;
	long			iTemp=0;
	
	for(i=0; i<iNumE; i++)
	{
	 /* if (	!giCL_Terrain || 
	  		(Terrain_Point(*(pV + (pE+i)->iPoint[0])) && Terrain_Point(*(pV + (pE+i)->iPoint[1]))))*/
		for(j=0; j<giNumObstWalls; j++)
		{
			sV_Tri = *(pV + (pE+i)->iPoint[0]);	/* the edges first point location */
			pW = gpsObst_Walls+j;
			iNumMatch=0;
			for(k=0; k<pW->iNumVert; k++)
			{
				sV_Wall = *(gpsObst_Vertices + *(pW->piVert + k));	
				if (Macro_PointsEqual2D(sV_Wall, sV_Tri) && sV_Wall.fz < 30)
				{
					iNumMatch++;
					if (iNumMatch == 1) {
						sV_Tri = *(pV + (pE+i)->iPoint[1]);	/* the edges second point location */
						k = 0;
					}
					else {
						k = pW->iNumVert;
						j = giNumObstWalls;
						(pE+i)->iWallType = pW->iWallType;
						(pE+i)->iThickness = pW->iThickness;
						iTemp++;
					}
				}
			}
		}
	}
}
/***************************************************************************************
	Assign_QP_to_Triangles:
		Associates each query point with a triangle by loading the info into the
  asTriQP array which is described at the start.
  	Uses the point location algorithm to get the Triangle ID and then simply adds
  the point to the current list for that ID.
***************************************************************************************/
void Assign_QP_to_Triangles()
{
	long			i, iTri, j;
	long*			piTemp;
	TRI_QP_TYPE*	psTri;

	gpsTriQP = (TRI_QP_TYPE*) calloc(gaiNum_Obst[2], sizeof(TRI_QP_TYPE));

	for(i=0; i < gaiNum_Obst[2]; i++)
	{
		(gpsTriQP+i)->iNum = 0;
		(gpsTriQP+i)->piTriQP = NULL;
	}

	for(i=0; i < giNumQP; i++)
	{
		iTri = PointLocation((gpsQPoints+i)->sLoc, gpObst_Tri, gaiNum_Obst, -1);

			/* Assign point to triangle */
	   	if (iTri >= 0) {
			psTri = (gpsTriQP + iTri);
			if (psTri->iNum == 0) 
				psTri->piTriQP = (long*) calloc(4, sizeof(long)); 
			else if (psTri->iNum % 4 == 0) {
				piTemp = psTri->piTriQP;
				psTri->piTriQP = (long*) calloc(psTri->iNum+4, sizeof(long));
				for(j=0; j<psTri->iNum; j++) 
					*(psTri->piTriQP+j) = *(piTemp+j);
				free(piTemp);
			}
			*(psTri->piTriQP + psTri->iNum) = i;
			psTri->iNum = psTri->iNum + 1;			
	   	}
	}
}

/***************************************************************************************
  Initialize_Data_Structures:
      Points_To_Tri - loads the points to be predicted into the triangle data 
           structure (Points.Tri[3])
      Obstacle_To_Tri - obstacle information into triangle data structure (Obst.Tri[3])
      Trans_Points - sets up the data structure for points to be considered by the 
           Prediction routine
***************************************************************************************/
void Initialize_Data_Structures()
{
  	int 	i;
  
  	QPoint_Cdt_Input_File();
  	i = system("../Triangle/triangle -pecY qpoints.cdt");
  	Setup_Tri_Data_Structure("qpoints.cdt.1.ele", gpQP_Tri, &gaiNumQP[2], &gaiNumQP[1]); 
  		/* Check_Tri_Data(gpQP_Tri, gaiNumQP, 1); */
 	Save_TriDS("QP_Tri.dat", gpQP_Tri, gaiNumQP); 
             /*************************************/
		/* creates a Geomview.OFF file from the Wall data */
	if (giCL_Geomview)	Geomview_Out("Wall_Data"); 
	  	/************ Builds tri data structure and adds wall info and slope *******/
		/************	 			to the edge array 						 *******/
  	Obst_Cdt_Input_File(); 
   	i = system("../Triangle/triangle -pecY obst.cdt");
    Setup_Tri_Data_Structure("obst.cdt.1.ele", gpObst_Tri, &gaiNum_Obst[2], &gaiNum_Obst[1]);
      /*Check_Tri_Data(gpObst_Tri, gaiNum_Obst, 0); */ 
	Edge_Real_or_Added(gpObst_Tri, gaiNum_Obst);
	  	/***************************************************************************/
	Save_TriDS("ObstTri.dat", gpObst_Tri, gaiNum_Obst);
	if (giCL_Geomview)	Geomview_Out("Obst_Tri"); 
}

/***************************************************************************************
				some data structure utility functions
***************************************************************************************/
long OtherTri(long iPrevTri, long iEdgeID)
{
	gpsEdge = (EDGE_TYPE*) gpObst_Tri[1];
	
	if (iEdgeID < 0)	return(-2);

	if ((gpsEdge+iEdgeID)->iTri[0] == iPrevTri)
		return((gpsEdge+iEdgeID)->iTri[1]);
	if ((gpsEdge+iEdgeID)->iTri[1] == iPrevTri)
		return((gpsEdge+iEdgeID)->iTri[0]);
	return(-2);
}
long Other_Edge(long iTriID, long iEdgeID1, long iEdgeID2)
{
	long iEdgeID3;

	if (iTriID < 0)    return(-1);

	gpsTri = (TRIANGLE_TYPE*) gpObst_Tri[2]; 
	if ((gpsTri+iTriID)->iEdge[0] != iEdgeID1 && (gpsTri+iTriID)->iEdge[0] != iEdgeID2) 
		iEdgeID3 = (gpsTri+iTriID)->iEdge[0]; 
	else if ((gpsTri+iTriID)->iEdge[1] != iEdgeID1 && (gpsTri+iTriID)->iEdge[1] != iEdgeID2) 
		iEdgeID3 = (gpsTri+iTriID)->iEdge[1]; 
	else 	
		iEdgeID3 = (gpsTri+iTriID)->iEdge[2];

	return(iEdgeID3);
}
		/* Find the vertex common to each edge and set sPCommon equal to it */
POINT_TYPE Common_Edge_Vertex(long iEdgeID1, long iEdgeID2) 
{
	POINT_TYPE	sPCommon;
	short		i;
 
	gpsVert = (VERTEX_TYPE*) gpObst_Tri[0]; gpsEdge = (EDGE_TYPE*) gpObst_Tri[1]; 
	if (Macro_PointsEqual3D(gpsVert[(gpsEdge+iEdgeID1)->iPoint[0]], gpsVert[(gpsEdge+iEdgeID2)->iPoint[0]]) ||  
			Macro_PointsEqual3D(gpsVert[(gpsEdge+iEdgeID1)->iPoint[0]], gpsVert[(gpsEdge+iEdgeID2)->iPoint[1]])) 
			i = 0;
	else 
			i = 1; 	
	sPCommon = gpsVert[(gpsEdge+iEdgeID1)->iPoint[i]]; 

	return(sPCommon);
}

long Tri_Edge_Segment_Inter(long iTriID, POINT_TYPE sP1, POINT_TYPE sP2)
{
	int				i;
	POINT_TYPE		sEdgePoint1, sEdgePoint2;
	long			Inter=-1, iEdgeID;

	gpsVert = (VERTEX_TYPE*) gpObst_Tri[0];
	gpsEdge = (EDGE_TYPE*) gpObst_Tri[1];
	gpsTri = (TRIANGLE_TYPE*) gpObst_Tri[2];

	i = 0; Inter = 0;
	while (i<3)
	{
		iEdgeID = (gpsTri+iTriID)->iEdge[i];
			/****/
		sEdgePoint1 = gpsVert[(gpsEdge+iEdgeID)->iPoint[0]];
		sEdgePoint2 = gpsVert[(gpsEdge+iEdgeID)->iPoint[1]];
		if (Intersect_Segments2D(sP1, sP2, sEdgePoint1, sEdgePoint2))
			{ i = 3; Inter = iEdgeID; }
		i++;
	}
	if (Inter < 0)	{ printf("Tri_Edge_Segment_Inter: no edge could be found"); cc=getchar(); }

	return(Inter);
}
