#include "formulas.h"
#include "predict.h"

/***************************************************************************************
								PATHLOSS COMPUTATIONS
***************************************************************************************/
/* float Reflection_Str(short iWallType, float fRayStr, float fSlope, short iFormula)
{
}

float Transmission_Str(short iWallType, float fRayStr, float fSlope, short iFormula)
{
}

float Scatter_Str(short iWallType, float fRayStr, float fSlope, short iFormula)
{
}

float Reflection_Slope(short iWallType, float fSlope, short iFormula)
{
}

float Transmission_Slope(short iWallType, float fApproachSlope)
{
}
   */
/***************************************************************************************
					PRE-CALCULATE/LOAD COEFFICIENTS
***************************************************************************************/
void Load_Wall_Coeff()
{
	short	i;
	float	fVal;

	giNumDielectric = 10;
	gafDielectric[0] = 1.0006;		/* Air */
	gafDielectric[1] = 15;			/* Ground */
	gafDielectric[2] = 5;			/* ceiling */
	gafDielectric[3] = 4.42;		/* Brick */
	gafDielectric[4] = 5;			/* Concrete  */
	gafDielectric[5] = 81;			/* Sea Water */
	gafDielectric[6] = 1.5;			/* Wood, lower, dry */
	gafDielectric[7] = 2.75;		/* Wood, middle, dry */
	gafDielectric[8] = 4;			/* Wood, upper, dry */
	gafDielectric[9] = 3.3;			/* Snow */
	gafDielectric[10] = 8;			/* Marble */

	for(i=0; i<giNumDielectric; i++)
		gafIndexOfRefraction[i] = (float) sqrt(gafDielectric[i]);

	for(i=0; i<=90; i++) {
		fVal = (float) (((float) i/ (float) 360) * PIE * 2);
		gafCos[i] = (float) fabs(cos(fVal));
		gafSin[i] = (float) sin(fVal);
	}
}
	/*
					  Formulas are taken from 
								
				Advanced Engineering Electromagnetics 
								by 
					  Constantine A. Balanis

		iPolarization = 0 -> Vertically polarized, else horizontally
	*/	
void Calc_Refl_Coeff(short iReflTransForm)
{
	short	i, j, k;
	double	fVal1, fVal2, fNumer, fDenom;

for(k=0; k<2; k++) /* polarization type */
{
	for(i=1; i<giNumDielectric; i++) 
	{
		for(j=0; j<=90; j++) 
		{
			if (k==1) {/* this formula assumes that the permeability of air and the
								wall are approxiamately equal page 187 5-18a */
				fVal1 = sqrt(gafDielectric[i]/gafDielectric[0]);
				fVal2 = 1 - (gafDielectric[0]/gafDielectric[i] * gafSin[j] * gafSin[j]);
				fVal2 = sqrt(fVal2);
				fVal1 = fVal1 * fVal2;
				fNumer = gafCos[j] - fVal1;
				fDenom = gafCos[j] + fVal1;
				gafReflAtten[k][i][j] = (float) fabs((fNumer / fDenom));
			}
			else {	/* page 191, 5-25a */
				fVal1 = sqrt(gafDielectric[0]/gafDielectric[i]);
				fVal2 = 1 - (gafDielectric[0]/gafDielectric[i] * gafSin[j] * gafSin[j]);
				fVal2 = sqrt(fVal2);
				fVal1 = fVal1 * fVal2;
				fNumer = fVal1 - gafCos[j];
				fDenom = gafCos[j] + fVal1;
				gafReflAtten[k][i][j] = (float) fabs((fNumer / fDenom));
			}
		} /* j - angle */
	} /* i - dielectric */
} /* k - polarization */
} 

void Calc_Trans_Coeff(short iReflTransForm)
{
	short	i, j, k;
	double	fVal1, fVal2, fNumer, fDenom;

for(k=0; k<2; k++) /* polarization type */
{
	for(i=1; i<giNumDielectric; i++) 
	{
		for(j=0; j<=90; j++) 
		{
			if (k==1) {/* this formula assumes that the permeability of air and the
								wall are approxiamately equal page 187 5-18a */
				fVal1 = sqrt(gafDielectric[i]/gafDielectric[0]);
				fVal2 = 1 - (gafDielectric[0]/gafDielectric[i] * gafSin[j] * gafSin[j]);
				fVal2 = fVal1 * sqrt(fVal2);
				fNumer = 2 * gafCos[j];
				fDenom = gafCos[j] + fVal2;
				gafTransAtten[k][i][j] = (float) fabs((fNumer / fDenom));
			}
			else {	/* page 191, 5-25a */
				fVal1 = sqrt(gafDielectric[0]/gafDielectric[i]);
				fVal2 = 1 - (gafDielectric[0]/gafDielectric[i] * gafSin[j] * gafSin[j]);
				fVal2 = fVal1 * sqrt(fVal2);
				fNumer = 2 * fVal1 * gafCos[j];
				fDenom = gafCos[j] + fVal2;
				gafTransAtten[k][i][j] = (float) fabs((fNumer / fDenom));
			}
		} /* j - angle */
	} /* i - dielectric */
} /* k - polarization */
} 
	/* fCosTheta is the cos of the angle of incidence */
	/* iType is the type of material - used for lookup in gafDieltric and gafIndexofRefraction */
	/* iPolar - determines whether the vertical (0) or horizontal (1) polarization formulas are used */
float Reflect_Attenuation(float fRayPow, short iType, short iPolar, float fCosTheta)
{
	float	fAttenuation;
	int		iAngle;

	iAngle = MyBinSearch(fCosTheta, gafCos, 90, 0);

	fAttenuation = gafReflAtten[iPolar][iType][iAngle];

	return(fAttenuation * fRayPow);
}

	/* fSinTheta is the sin of the angle of incidence */
	/* iType is the type of material - used for lookup in gafDieltric and gafIndexofRefraction */
	/* iPolar - determines whether the vertical (0) or horizontal (1) polarization formulas are used */
float Trans_Attenuation(float fRayPow, short iType, short iPolar, float fSinTheta)
{
	float	fAttenuation, fVal;
	int		iAngle;

	fVal = gafIndexOfRefraction[iType] * fSinTheta;

	iAngle = MyBinSearch(fVal, gafSin, 90, 0);

	fAttenuation = gafTransAtten[iPolar][iType][iAngle];

	return(fAttenuation * fRayPow);
}

