#ifndef FORMULAS

#define FORMULAS

#define 	POWER_THRESHOLD_HEAR	1e-9	 /* -90dBm */
#define		PIE						3.1415927

/**************************************************************************************
	Free_Space_Calc:

		Friis equation is used:

			Pr/Pt = Gt * Gr (c / 4 * (pie) * f * d) ^ 2

			where c = speed of signal through medium
				  f = radio signal frequency
				  d = distance traveled
				  Pr = signal power at receiver
				  Pt = signal power at transmitter
		- Assumes the attena is omnidirectional so the gain of the transmitter = 1
		- Assumes teh receiver " " so gain = 1	

		so we get

			Pr = Pt * (c / 4 * pie * f * d) ^ 2

			but gfFree = (c / 4 * pie * f) ^ 2 so the final formula is

			Pr = Pt * gfFree * (1 / d^2)

     sT - the transmitter location in a point type
     sD - the destination location in a point type
****************************************************************************************/
#define Macro_Free_Space_Calc(fTransPow, fDSquared)  (float) (fTransPow * gfFree / fDSquared)

#define Square(x)      x * x

/***************************************************************************************
                           Global Variables
***************************************************************************************/

double			gafDielectric[20];
int				giNumDielectric;
float			gafCos[91], gafSin[91];
float			gafReflAtten[2][20][91];
float			gafTransAtten[2][20][91];
float			gafIndexOfRefraction[20];

void Load_Wall_Coeff();
void Calc_Refl_Coeff(short iFormula);
void Calc_Trans_Coeff(short iFormula);
float Reflect_Attenuation(float fRayPow, short iType, short iPolar, float fCosTheta);
float Trans_Attenuation(float fRayPow, short iType, short iPolar, float fCosTheta);

#endif



