/**************************************************************************************
	The following functions load the user supplied data:
		- wall information (obst.dat)
		- transmitter info (transmitter.dat)
		- query point locations (qpoints.dat)
		- user definable parameters (world.dat; not yet implemented)
**************************************************************************************/

#include "predict.h"

/**************************************************************************************
  QPoint_Data_Type:  This function loads the non-uniform set of query points into
        an array.  Returns the pointer to the head of the array.
**************************************************************************************/
VERTEX_TYPE* Load_QPoints_To_Predict()
{
  FILE              *fpin;
  long              j;
  QP_TYPE* 			psQPoints, *p;
  VERTEX_TYPE*		pVert;

  fpin=fopen("qpoints.dat","r");
  if (fpin==NULL)    ERROR_Message("Error opening file: ", "qpoints.dat"); 

  fscanf(fpin,"%li\n",&giNumQP);
               /* allocate space for the query point array */
  psQPoints = (QP_TYPE*) calloc(giNumQP, sizeof(QP_TYPE));
  				/* used to store Vertex info in the Tri data structure */
  gaiNumQP[0] = giNumQP;
  pVert = (VERTEX_TYPE*) calloc(gaiNumQP[0], sizeof(VERTEX_TYPE));
  
  p = psQPoints;
  for(j=0;j<giNumQP;j++)
  {
    if (!gi3DOn) 
      fscanf(fpin,"%f %f\n",&(p->sLoc.fx),&(p->sLoc.fy));
    else
      fscanf(fpin,"%f %f %f\n",&(p->sLoc.fx),&(p->sLoc.fy),&(p->sLoc.fz));
	if (p->sLoc.fz < 0)	
		ERROR_Message("Load_QPoints", "query point data contains negative z value!");
  	p->iBest_Trans_ID = -1;
  	p->fBest_Path_Loss = (float) -300;
  	p->iNum_Trans_Can_Hear = 0;
  	p->psSignal_Data = NULL;
    p++;
  }
  fclose(fpin);
  	
  for(j=0;j<giNumQP;j++)
  	{
		(pVert+j)->fx = (psQPoints+j)->sLoc.fx;
		(pVert+j)->fy = (psQPoints+j)->sLoc.fy;
		(pVert+j)->fz = (psQPoints+j)->sLoc.fz;
	}
		/* add the bounding region at the end for use in the cdt output 
  (pVert+j)->fx = gasBR[0].fx; (pVert+j)->fy = gasBR[0].fy; (pVert+j)->fz = (float) 0;
  (pVert+j+1)->fx = gasBR[0].fx; (pVert+j+1)->fy = gasBR[1].fy; (pVert+j+1)->fz= (float) 0; 
  (pVert+j+2)->fx = gasBR[1].fx; (pVert+j+2)->fy = gasBR[0].fy; (pVert+j+2)->fz= (float) 0; 
  (pVert+j+3)->fx = gasBR[1].fx; (pVert+j+3)->fy = gasBR[1].fy; (pVert+j+3)->fz = (float) 0; */

  ERROR_Check_For_Duplicate_Vertices(pVert, gaiNumQP[0], "qpoints.dat");
  
  gpsQPoints = psQPoints;	

  return(pVert);
}


/**************************************************************************************
  Load_Obstacles:  This loads in the building information.  The file 
      assumes a particular format (see... for more details).
	This works for a flat terrain and the typical building.

	  NOTE: might be a problem matching the end points of walls up.  Just possibly,
	  			not probably.
**************************************************************************************/
WALL_TYPE* Load_Obstacles(VERTEX_TYPE** psVert)
{  
  FILE              *fpin;
  int               k, iFirst=1;
  long              i;
  WALL_TYPE*     	psWalls, *pW;
  VERTEX_TYPE*		pV;
  long*				piVert;
  float				fMinX, fMinY, fMaxX, fMaxY, fMaxZ;

  fpin=fopen("obst.dat","r");
  if (fpin==NULL)    ERROR_Message("Error opening file: ", "obst.dat"); 

  fscanf(fpin,"%li %li\n", &giNumObstVert, &giNumObstWalls);	/* number of vertices and walls */
  
         /* allocate space for building wall points array (upper and lower corners) */
  *psVert = (VERTEX_TYPE*) calloc(giNumObstVert+4, sizeof(VERTEX_TYPE));
  psWalls = (WALL_TYPE*) calloc(giNumObstWalls, sizeof(WALL_TYPE));
  pW = psWalls;
  pV = *psVert;

  	/* Load the vertice information */
  for(i=0;i<giNumObstVert;i++)
  {
    if (!gi3DOn)
     fscanf(fpin,"%f %f\n",&(pV->fx),&(pV->fy));
    else
     fscanf(fpin,"%f %f %f\n", &(pV->fx), &(pV->fy), &(pV->fz));
	if (pV->fz < 0)	
		ERROR_Message("Load_Obstacles", "obstacle data contains negative z value!");
	 	/* this is for the computation of the rectangular bounding region */
	if (iFirst) {  fMinX=pV->fx; fMinY=pV->fy; fMaxX=pV->fx; fMaxY=pV->fy; fMaxZ=pV->fz; iFirst=0; }
	if (pV->fx < fMinX)		fMinX = pV->fx;
	if (pV->fy < fMinY)		fMinY = pV->fy;
	if (pV->fx > fMaxX)		fMaxX = pV->fx;
	if (pV->fy > fMaxY)		fMaxY = pV->fy;
	if (pV->fz > fMaxZ)		fMaxZ = pV->fz;
    pV++;
  }
    	/* boundary */
  gasBR[0].fx = fMinX - (float) 40;  gasBR[0].fy = fMinY - (float) 40; gasBR[0].fz = (float) 0;
  gasBR[1].fx = fMaxX + (float) 40;  gasBR[1].fy = fMaxY + (float) 40; gasBR[1].fz = fMaxZ;

		/* add the bounding region at the end for use in the cdt output */
  pV->fx = gasBR[0].fx; pV->fy = gasBR[0].fy; pV->fz=(float) 0; pV++;
  pV->fx = gasBR[0].fx; pV->fy = gasBR[1].fy; pV->fz=(float) 0; pV++;
  pV->fx = gasBR[1].fx; pV->fy = gasBR[0].fy; pV->fz=(float) 0; pV++;
  pV->fx = gasBR[1].fx; pV->fy = gasBR[1].fy; pV->fz=(float) 0;
  giNumObstVert = giNumObstVert+4; /* for boundary points */

  ERROR_Check_For_Duplicate_Vertices(*psVert, giNumObstVert, "obst.dat");

  	/* Load the wall information */
  for(i=0;i<giNumObstWalls;i++)
  {
	fscanf(fpin, "%hi", &(pW->iNumVert));
   			/* allocate memory for storage of the vertice IDS */
	piVert = (long*) calloc(pW->iNumVert, sizeof(long)); 
	pW->piVert = piVert;

   	for(k=0; k < pW->iNumVert; k++)
   	{ 
		fscanf(fpin, "%li", &(*(piVert+k)));			/* store the value */
                *(piVert+k) = *(piVert+k) - 1; /* fix obst.dat and then rermove this */
	}
      
    ERROR_Check_Verts_Planar(piVert, pW->iNumVert);

    fscanf(fpin,"%hi %hi\n", &(pW->iWallType), &(pW->iThickness));
	pW++;
      }

  fclose(fpin); 

 return(psWalls);
}
  
  
/**************************************************************************************
  Load_Transmitter_Info:  This loads the transmitter information into an array of
    transmitter structures.  Returns pointer to the head of the array
***************************************************************************************/
TRANS_TYPE* Load_Transmitter_Info()
{
  FILE                *fpin;
  int                 i;
  TRANS_TYPE*    psTrans, *p;
 
  fpin=fopen("transmitter.dat","r");
  if (fpin==NULL)    ERROR_Message("Error opening file: ", "transmitter.dat"); 

  fscanf(fpin,"%hi\n", &giNum_Trans);

  psTrans = (TRANS_TYPE*) calloc(giNum_Trans, sizeof(TRANS_TYPE));
  p = psTrans;
  
  for(i=0;i<giNum_Trans;i++)
  {
    fscanf(fpin,"%f %f %f %f %f %f %hi %hi %f %f %hi\n",
           &(p->sLoc.fx),
           &(p->sLoc.fy),
           &(p->sLoc.fz),
           &(p->fFreq_LRange),
           &(p->fFreq_HRange),
           &(p->fMax_Power),
           &(p->iModel_Type),
           &(p->iOmni_directional),
           &(p->fAng1),
           &(p->fAng2),
           &(p->iPolarization));
    p->piQPoints = NULL;
    p->iNum_Points = 0;
    p->iTrans_ID = i;
	if (p->iPolarization != 0 && p->iPolarization != 1) { p->iPolarization = 1; }
  
    p++;
  }
  
  fclose(fpin);

  return(psTrans);
}

/**************************************************************************************	
	Load_Terrain:  loads the terrain height field data from TERRAIN.DAT
		- first line is the number of points
		- each line is a (x,y,z) point 
			x y z (separated by a space)
	* this file must contain the EXACT points as those used for the base points
		for walls.  The wall points are then matched to the height field data
		to decide which edges are needed in the triangulation.
**************************************************************************************/
POINT_TYPE* Load_Terrain()
{
  	FILE              	*fpin;
  	long              	j;
 	POINT_TYPE*			psPoints, *psP;
	long   iCount=0;

  	fpin=fopen("terrain.dat","r");
  	if (fpin==NULL)    ERROR_Message("Error opening file: ", "terrain.dat"); 

  	fscanf(fpin,"%li\n", &giNumTerr);
 		/* allocate space for the query point array */
  	psPoints = (POINT_TYPE*) calloc(giNumTerr, sizeof(POINT_TYPE));
  
  	psP = psPoints;
  	for(j=0;j<giNumTerr;j++)
  	{
    	if (!gi3DOn) 
      		fscanf(fpin,"%f %f\n",&(psP->fx),&(psP->fy));
    	else
      		fscanf(fpin,"%f %f %f\n",&(psP->fx),&(psP->fy),&(psP->fz));
		if (psP->fz < 0)	
			ERROR_Message("Load_Terrain", "heigh field data contains negative z value!");
	 psP++; 
  	}
  	fclose(fpin);
 	
  	ERROR_Check_For_Duplicate_Vertices(psPoints, giNumTerr, "terrain.dat");
  
	
  	return(psPoints);
}

/***************************************************************************************
  Load_Data:
      Load_QPoints_To_Predict - loads the points to be predicted into the points array
      Load_Buildings - loads building information 
      Load_Transmitter_Info - Loads transmitter information into memory.
           Returns pointer to top of array that holds pointers to each struct for a
	   transmitter.  
***************************************************************************************/
void Load_Data()
{
  	if (giCL_Terrain)		gpsTerrain = Load_Terrain();
  	gpsObst_Walls = Load_Obstacles(&gpsObst_Vertices);
  	gpQP_Tri[0] = (void*) Load_QPoints_To_Predict();
  	gpsTrans_Data = Load_Transmitter_Info();

		/* a global 'constant' used in the Friis free space calculation */
	gfC = (float) 3e8;	/* speed of light */
}


/***************************************************************************************
  Get_Assumptions:
      Loads the user specified assumptions from the assump.txt file and determines the
    other assumptions from selected prediction models.
****************************************************************************************/
void Get_Assumptions()
{
  gi3DOn = 1;
  giObst = 1;
}

