
#include "predict.h"
#include "formulas.h"

/**************************************************************************************
 	Add_Signal_Data: 
		Adds the predictions made by the selected method to the Query Point 
	data structure.
****************************************************************************************/
void Add_Signal_Data(QP_TYPE* psQP, TRANS_TYPE sT, float fPower, 
					float fShortPath, float fShortStr, float fLongPath, float fLongStr)
{
  	SIGNAL_TYPE*  	psSignal;
  	float			fPath_Loss = (float) 10 * (float) log10(fPower);
/*  	float			fDelay_Spread = (fLongPath - fShortPath) / gfC; */
  	float			fDelay_Spread = (fLongPath - fShortPath) / gfC;
  	short   		n;

	if (fPower >= POWER_THRESHOLD_HEAR)
  	{
	  	if (fPath_Loss > psQP->fBest_Path_Loss) {
			psQP->fBest_Path_Loss = fPath_Loss;
			psQP->iBest_Trans_ID = sT.iTrans_ID;
		}
	    if (psQP->iNum_Trans_Can_Hear == 0) {
		  psSignal = (SIGNAL_TYPE*) calloc(1, sizeof(SIGNAL_TYPE));
		  psQP->psSignal_Data = psSignal;
		}
	    else {
		  psSignal = psQP->psSignal_Data;
		  for(n=1; n<psQP->iNum_Trans_Can_Hear; n++)
		    psSignal = psSignal->psNext;
		  psSignal->psNext = (SIGNAL_TYPE*) calloc(1, sizeof(SIGNAL_TYPE));
		  psSignal = psSignal->psNext;
		}
	    psSignal->fPath_Loss = fPath_Loss;
	    psSignal->fDelay_Spread = fDelay_Spread;
	    psSignal->iTrans_ID = sT.iTrans_ID;	  
		psSignal->fShortPath = fShortPath;
		psSignal->fShortStr = fShortStr;
	    psSignal->psNext = NULL;
	    psQP->iNum_Trans_Can_Hear = psQP->iNum_Trans_Can_Hear + 1;
    }
}

/**************************************************************************************
  Model_Free_Space:
  		This is the standard Friis free space model
****************************************************************************************/
void Model_Free_Space(TRANS_TYPE sT)
{
  QP_TYPE* 		psQP = gpsQPoints, sQPcurr;
  long*       	piTP = sT.piQPoints;
  int          	i;
  float        	fPower, fDist;
  
  for(i=0; i< sT.iNum_Points; i++)
    {
        sQPcurr = *(psQP + *(piTP + i));
        fDist = Macro_Dist(sT.sLoc, sQPcurr.sLoc); 
  		fPower = Macro_Free_Space_Calc(sT.fMax_Power, fDist); 
      	Add_Signal_Data(psQP + *(piTP + i), sT, fPower, fDist, fPower, fDist, fPower); 
    }
}

/**************************************************************************************
  Predict_Points:  
     Trans_Points - loads each transmitter with it's associated points to be 
     		predicted upon
       
       Loops over all transmitters calling the appropriate method for prediction for 
      each transmitter.
****************************************************************************************/
void Predict_Points()
{
  	TRANS_TYPE* 	psTrans = gpsTrans_Data;
  	short  			t, i;
  
  	for(t=0; t<giNum_Trans; t++)
    {
      	gfFree = Square(gfC / (4 * (float) PIE * (psTrans->fFreq_HRange)));

      	switch (psTrans->iModel_Type) 
		{
			case 1: /* Ray_Method();*/
				i = Ray_Method(t, giCL_max_events, gfCL_beta_threshold, (short) 1, (short) 1);
				break;
			default:
	  			Model_Free_Space(*psTrans);
			break;
		} 
      	psTrans++;
    }  
}
/*********************************************************************
  Process_Command_Line:
		Checks each command line argument and then sets the appropriate
	global variables.
**********************************************************************/
void Process_Command_Line(int argc, char* argv[])
{
	int 		i;
	char 		c = ' ', s[] = "         ";
		/* default settings */
	giCL_Initialize = 0;
	giCL_recursion_level = 3;
	giCL_max_events = 3;
	gfCL_beta_threshold =pow(10, -8.75);
	giCL_Split = 0;
	giCL_RaySave = 0;
	giCL_Geomview = 0;
	giCL_Debug = 0; gfp_Debug = NULL; 
	giCL_Stats = 1;						/* print run statistics */
	gfCL_Split = (float) 10;				/* maximum area for a cone end before being split */
	giCL_InBuilding = 0;
	giCL_Terrain = 0;
		
	for(i=1; i<argc; i++)
	{
		if (argv[i][0] == '?') {
			printf("\nThe following are PREDICT's command line arguments:\n\n");
			printf(" i - initialize: use this option when one is running a scenario\n	for the very first time\n");
			printf("n# - number of recursions during tesselation of the sphere (1-5)\n");
			printf("	this affects the number of initial rays launched\n");
			printf("e# - set the maximum number of events that a ray can undergo (0-9)\n");
			printf("s  - splitting of launched rays is off\n");
			printf("a  - enter split area value (see manual)\n");
			printf("		Ex: Predict a3.4   sets the split area to 3.4\n");
			printf("f  - raysave.dat file created\n");
			printf("g  - geomview files are created: ObstTri.OFF, QPTri.OFF, Walls.OFF\n");
			printf("d  - debugging is on: the system will now self-check itself.\n");
			printf("D  - debugging is on and a file Debug.txt will be created.\n");
			printf("t  - turn statistical output off\n");
			printf("b# - set beta threshold: this value sets the signal cut threshold\n");
			printf("		to 10^-#, which is equivalent to 10 * -# dBm.\n");
			printf("I - this sets PREDICT to run using in-building options\n");
			printf("T - tells PREDICT to load the user provide height field data\n");
			exit(1);
		}
		else if (argv[i][0] == 'n') {
			c = argv[i][1];  
			giCL_recursion_level = atoi(&c); 
			if (giCL_recursion_level < 1)		giCL_recursion_level = 1;
			if (giCL_recursion_level > 5)		giCL_recursion_level = 5;
		}
		else if (argv[i][0] == 'b') {
			c = argv[i][1];  
			strcpy(s, &argv[i][1]);
			gfCL_beta_threshold = pow(10, -1 * atof(s)); }
		else if (argv[i][0] == 'e') {
			c = argv[i][1];  
			giCL_max_events = atoi(&c); }
		else if (argv[i][0] == 's') 
			giCL_Split = 0;
		else if (argv[i][0] == 'i') 
			giCL_Initialize = 1;
		else if (argv[i][0] == 'f') 
			giCL_RaySave = 1;
		else if (argv[i][0] == 'g') 
			giCL_Geomview = 1;
		else if (argv[i][0] == 't') 
			giCL_Stats = 0;
		else if (argv[i][0] == 'T') 
			giCL_Terrain = 1;
		else if (argv[i][0] == 'I') 
			giCL_InBuilding = 1;
		else if (argv[i][0] == 'a') {
			strcpy(s, &argv[i][1]);
			gfCL_Split = (float) atof(s);
		}
		else if (argv[i][0] == 'P') {
			giCL_SavePaths = 1;
		}
		else if (argv[i][0] == 'd') 
			giCL_Debug = 1; 
		else if (argv[i][0] == 'D') {
			giCL_Debug = 2; 
			OpenDF(); }
	}
	if (giCL_InBuilding && giCL_Terrain) {
		printf("\nCannot provide height field data when in in-building mode\n");
		exit(1);
	}
	if (giCL_InBuilding)
		printf("PREDICT is set for IN-BUILDING prediction\n");
	else {
		printf("PREDICT is set for OUTDOOR prediction\n");
		if (giCL_Terrain)	printf("	and is loading height field data\n");
	}
	printf("Maximum # of allowed events: %i\n", giCL_max_events);
	printf("Recursion level for tesselation of sphere: %i\n", giCL_recursion_level);
	printf("Beta threshold is set at %2.1f dBm\n", (int) 10 * log10(gfCL_beta_threshold));
	if (giCL_RaySave == 1) 	printf("Raysave file will be created\n");
	if (giCL_Geomview == 1) printf("Geomview files will be created\n");
	if (giCL_Debug > 0)	printf("DEBUGGING IS ON\n");
	if (giCL_Split == 1)	printf("Cone Splitting is active at area %f\n", gfCL_Split);
	
	gfp_Paths = fopen("paths.dat", "w");

	printf("\n\n");
	gfCL_Split = (float) pow(gfCL_Split * 2, 2);
}

main(argc, argv)
int             argc;
char            *argv[]; 
{
		/*********************************************************************
				Set up system to run based upon command codes
		**********************************************************************/
	Process_Command_Line(argc, argv);
		/*********************************************************************
				Load user defined assumptions and the current scenario
		**********************************************************************/
  	Get_Assumptions();
  	Load_Data();
		/* load raw data, run TRIANGLE on it and create data structures */
	if (giCL_Initialize) {
		Initialize_Data_Structures();   }
	else {	/* TRIANGLE has already been run so skip creation and just load the data */
		Load_TriDS("ObstTri.dat", gpObst_Tri, gaiNum_Obst);
   		Load_TriDS("QP_Tri.dat", gpQP_Tri, gaiNumQP); 
	}
		/*************************************/

  	Compute_Trans_QPoints(); 
	Assign_QP_to_Triangles();
	if (giCL_Terrain)	gpsTerrPlanes = Compute_Terrain_Tri_Planes();

		/*************** Not needed anymore **********************/
	free(gpsObst_Walls); giNumObstVert = 0; gpsObst_Walls = NULL;
	free(gpsObst_Vertices); giNumObstWalls = 0; gpsObst_Vertices = NULL;
	if (giCL_Terrain) { free(gpsTerrain); giNumTerr = 0; }

		/*************** Make Signal Predictions **********************/
	Predict_Points(); 
	printf("Done predict"); cc=getchar();

	CloseDF();	/* for debugging purposes */

	Save_Query_Points(gpsQPoints, gpQP_Tri, gaiNumQP);

	fclose(gfp_Paths); /* for paths.dat file */

 return(1); 
}


