/***************************************************************************************
							UTILITY FUNCTIONS
***************************************************************************************/
#include "predict.h"

	/* rotating through the yz or xz axis */
POINT_TYPE Rotate_CosXYZ(POINT_TYPE sP, float fCosTheta, float fSinTheta, short iRot)
{
	POINT_TYPE	sR;
	float		fVal1, fVal2;

	if (iRot == 0) {
		sR.fx = sP.fx;
		fVal1 = fCosTheta * sP.fy - fSinTheta * sP.fz;
		fVal2 = fSinTheta * sP.fy + fCosTheta * sP.fz;
		sR.fy = fVal1;  sR.fz = fVal2;
	}
	else {
		sR.fy = sP.fy;
		fVal1 = fCosTheta * sP.fx - fSinTheta * sP.fz;
		fVal2 = fSinTheta * sP.fx + fCosTheta * sP.fz;
		sR.fx = fVal1;  sR.fz = fVal2;
	}

	return(sR);
}

VECTOR_TYPE Vector_MultBy_Scaler(VECTOR_TYPE sV, float v)	
{
	sV.i = sV.i * v; 
	sV.j = sV.j * v; 
	sV.k = sV.k * v;

	return(sV);
}

POINT_TYPE MidPoint(POINT_TYPE p1, POINT_TYPE p2) 
{
	POINT_TYPE	midpoint;

	midpoint.fx = ( p1.fx + p2.fx ) / (float) 2.0; 
    midpoint.fy = ( p1.fy + p2.fy ) / (float) 2.0; 
    midpoint.fz = ( p1.fz + p2.fz ) / (float) 2.0;

	return(midpoint);
}
	
POINT_TYPE ParaLine_Calc_Point(VECTOR_TYPE sU, POINT_TYPE sO, float t) 
{
	POINT_TYPE	sP;
	
	sP.fx = sO.fx + t * sU.i; 
	sP.fy = sO.fy + t * sU.j; 
	sP.fz = sO.fz + t * sU.k;

	return(sP);
}

/**********************************************************************/
/** return twice the area of a triangle with vertices A,B,c **/

float Area2( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC )
{
/*
  printf("\n%f %f %f %f %f %f\n", sA.fx, sA.fy, sB.fx, sB.fy, sC.fx, sC.fy);
*/  

    return 
        sA.fx * sB.fy - sA.fy * sB.fx 
	    + sA.fy * sC.fx - sA.fx * sC.fy
		+ sB.fx * sC.fy - sB.fy * sC.fx;
}

/***************************************************************************/

bool XOR( bool x, bool y )
{
    return !x ^ !y;
}

/***************************************************************************/

bool Left( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC )
{
  return Area2( sA, sB, sC ) > 0.000000 ;
}

/***************************************************************************/
/**
  return t if sA, sB, sC are collinear and sC lies between the closed
  segment sAsB
**/
bool Between( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC )
{
    if( ! Collinear( sA,sB,sC ))
	return 0;
    
    /** if ab not vertical, check betweenness on x; else on y **/
    if( sA.fx != sB.fx )
	{
	    return 
		( (sA.fx <= sC.fx) && (sC.fx <= sB.fx) ) ||
		    ( (sA.fx >= sC.fx) && (sC.fx >= sB.fx) );
	}
    else 
	{
	    return
		( (sA.fy <= sC.fy) && (sC.fy <= sB.fy) ) ||
		    ( (sA.fy >= sC.fy) && (sC.fy >= sB.fy) );
	}
}

/***************************************************************************/
/**
return 1 if the given point is inside the triangle
**/
bool InsideTri( POINT_TYPE sP, TRIANGLE_TYPE sTriangle )
{
    POINT_TYPE p1, p2, p3;
    
	gpsVert = (VERTEX_TYPE*) gpObst_Tri[0];

    p1 = *(gpsVert + sTriangle.iPoint[0]);
    p2 = *(gpsVert + sTriangle.iPoint[1]);
    p3 = *(gpsVert + sTriangle.iPoint[2]);

    if( Between(p1,p2,sP) || Between(p2,p3,sP) || Between(p3,p1,sP) ||
       (Left(p1,p2,sP) && Left(p2,p3,sP) && Left(p3,p1,sP)) ||
       ((! Left(p1,p2,sP)) && (! Left(p2,p3,sP)) && (! Left(p3,p1,sP))))
	{
	    return 1;
	}
    else
	{
	    return 0;
	}
}
bool Inside_Tri( POINT_TYPE sP, POINT_TYPE p1, POINT_TYPE p2, POINT_TYPE p3)
{
    if( Between(p1,p2,sP) || Between(p2,p3,sP) || Between(p3,p1,sP) ||
       (Left(p1,p2,sP) && Left(p2,p3,sP) && Left(p3,p1,sP)) ||
       ((! Left(p1,p2,sP)) && (! Left(p2,p3,sP)) && (! Left(p3,p1,sP))))
	{
	    return 1;
	}
    else
	{
	    return 0;
	}
}

/***************************************************************************/

bool Collinear( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC )
{
    float A = Area2( sA, sB, sC );

    return (A > -0.00001) && (A < 0.00001);
}

/***************************************************************************/
/**
return t if AB lies on CD (they could be the same segment)
**/

bool LieOn( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC, POINT_TYPE sD )
{
    return Between(sC,sD,sA) && Between(sC,sD,sB);
}


/***************************************************************************/

bool IntersectProp( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC, POINT_TYPE sD )
{
    if( Collinear( sA, sB, sC ) ||
       Collinear( sA, sB, sD ) ||
       Collinear( sC, sD, sA ) ||
       Collinear( sC, sD, sB ))
	return 0;

  return 
      ( XOR( Left(sA,sB,sC), Left(sA,sB,sD)) 
       && XOR( Left(sC,sD,sA), Left(sC,sD,sB)));
}

/***************************************************************************/

bool Intersect_Segments2D( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC, POINT_TYPE sD )
{
    if( IntersectProp( sA,sB,sC,sD ))
	{
	    return 1;
	}
    else if(   Between(sA,sB,sC) || Between(sA,sB,sD) 
	    || Between(sC,sD,sA) || Between(sC,sD,sB) )
	{
	    return 1;
	}
    else
	{
	    return 0;
	}
}

UNIT_VECTOR Normalize_Vector(VECTOR_TYPE sV)	
{
	float	fV = (float) sqrt(sV.i * sV.i + sV.j * sV.j + sV.k * sV.k); 

	sV.i = sV.i / fV;
	sV.j = sV.j / fV;
	sV.k = sV.k / fV;

	return(sV);
}

/***************************************************************************
	The vector points in the direction P1 --> P2
/***************************************************************************/
PARA_LINE_TYPE	Line_Parametric_Equation3D(POINT_TYPE sP1, POINT_TYPE sP2)
{
	PARA_LINE_TYPE	sPara;
	
	sPara.sV.i = sP2.fx - sP1.fx;
	sPara.sV.j = sP2.fy - sP1.fy;	
	sPara.sV.k = sP2.fz - sP1.fz;

	if (giCL_Debug && Macro_Zero(sPara.sV.i) && Macro_Zero(sPara.sV.j) && Macro_Zero(sPara.sV.k))	{
		fPrint_Point(gfp_Debug, sP1); fPrint_Point(gfp_Debug, sP2);
		fprintf(gfp_Debug, "Line_Parametric_Equation3D sP1 == sP2");
	}

	sPara.sP = sP1;
	sPara.sV = Normalize_Vector(sPara.sV);

 	return(sPara);
}
/***************************************************************************
	computes the normal to the plane defined by the three input points and
	then makes it point towards the side the provided point lies upon
***************************************************************************/
PLANE_TYPE Plane_Point_Normal_Plus(POINT_TYPE sP1, POINT_TYPE sP2, 
														POINT_TYPE sP3, POINT_TYPE sPdir)
{
	UNIT_VECTOR		sU = Plane_Point_Normal(sP1, sP2, sP3);
	PLANE_TYPE		sPlane;

	if (Point_To_Plane_DistancePM2(sU, sP1, sPdir) < 0)
			{ sU = Vector_MultBy_Scaler(sU, (float) -1); }

	sPlane.sN = sU;	sPlane.sP = sP1;

	return(sPlane);
}
/***************************************************************************
	computes the normal to the plane defined by the three input points and
	then makes it point away from the side the provided point lies upon
***************************************************************************/
UNIT_VECTOR	Plane_Point_Normal_Minus(POINT_TYPE sP1, POINT_TYPE sP2, 
														POINT_TYPE sP3, POINT_TYPE sPdir)
{
	UNIT_VECTOR		sU = Plane_Point_Normal(sP1, sP2, sP3);

	if (Point_To_Plane_DistancePM2(sU, sP1, sPdir) > 0)
			{ sU = Vector_MultBy_Scaler(sU, (float) -1); }

	return(sU);
}

/***************************************************************************
	computes the normal to the plane defined by the three input points
***************************************************************************/
UNIT_VECTOR	Plane_Point_Normal(POINT_TYPE sP1, POINT_TYPE sP2, POINT_TYPE sP3)
{
	float			fXDiff[2], fYDiff[2], fZDiff[2], fX, fY, fZ;
	float			fC;
	VECTOR_TYPE		sV;
	POINT_TYPE		sP;
	short			l;
	PLANE_TYPE		sPlane;
	PARA_LINE_TYPE	sLine1, sLine2; 

	if (Macro_PointsEqual3D(sP1, sP2) || Macro_PointsEqual3D(sP1, sP3) || Macro_PointsEqual3D(sP2, sP3)) {
		if (giCL_Debug) {
			Print_Point(sP1); Print_Point(sP2); Print_Point(sP3);
			printf("Plane_Point_Normal: sP1, sP2, sP3 are not distinct"); 
		}
		sPlane.sN.i = (float) 0; sPlane.sN.j = (float) 0; sPlane.sN.k = (float) 1;			   
		return(sPlane.sN);
	}
	sLine1 = Line_Parametric_Equation3D(sP1, sP2);
	sLine2 = Line_Parametric_Equation3D(sP1, sP3);
	if (sLine1.sV.i == sLine2.sV.i && sLine1.sV.j == sLine2.sV.j && sLine1.sV.k == sLine2.sV.k) {
		Print_Point(sP1); Print_Point(sP2); Print_Point(sP3);  
		return(sLine1.sV);
	}
		/* for the test, remove later */
	for(l=0; l<2; l++) { fXDiff[l] = (float) 0; fYDiff[l]=(float) 0; fZDiff[l]=(float) 0;}
		
	if (Macro_FloatEqual(sP1.fx, sP2.fx) && Macro_FloatEqual(sP1.fx, sP3.fx))	
		{ sV.i = (float) 1; sV.j = (float) 0; sV.k = (float) 0; }
	else if (Macro_FloatEqual(sP1.fy, sP2.fy) && Macro_FloatEqual(sP1.fy, sP3.fy))	
		{ sV.i = (float) 0; sV.j = (float) 1; sV.k = (float) 0; }
	else if (Macro_FloatEqual(sP1.fz, sP2.fz) && Macro_FloatEqual(sP1.fz, sP3.fz))	
		{ sV.i = (float) 0; sV.j = (float) 0; sV.k = (float) 1; }
	else {
		for(l=0; l<2; l++) {
			if (l == 0)	sP = sP2;	else sP = sP3;
			fXDiff[l] = sP.fx - sP1.fx; 
			fYDiff[l] = sP.fy - sP1.fy; 
			fZDiff[l] = sP.fz - sP1.fz; 
		}

		if (fabs(fXDiff[0]) < fabs(fXDiff[1]))
		{	
			fX = fXDiff[0]; fY = fYDiff[0]; fZ = fZDiff[0];
			fXDiff[0] = fXDiff[1]; fYDiff[0] = fYDiff[1]; fZDiff[0] = fZDiff[1];
			fXDiff[1] = fX; fYDiff[1] = fY; fZDiff[1] = fZ;
		}

		fC = fXDiff[1] / fXDiff[0];		/* performing one step of gaussian elimination */ 
		fXDiff[1] = (float) 0; fYDiff[1] = fYDiff[1] - fYDiff[0] * fC; fZDiff[1] = fZDiff[1] - fZDiff[0] * fC; 
	
		
		sV.k = (float) 1;
		if (Macro_NotZero(fYDiff[1]))
			sV.j = ((float) -1 * fZDiff[1] * sV.k) / fYDiff[1];		
		else 
			{ sV.j = (float) 1; sV.k = (float) 0; }
		sV.i = ((float) -1 * (fZDiff[0] * sV.k + fYDiff[0] * sV.j)) / fXDiff[0];

	 	sV = Normalize_Vector(sV);
	}
	sPlane.sN = sV;
	sPlane.sP = sP1;

		/* a validation test for numerical roundoff errors */
if (giCL_Debug) {
	sV.i = fXDiff[0]; sV.j = fYDiff[0]; sV.k = fZDiff[0];
	fC = Macro_DotProduct(sV, sPlane.sN);
	if (Macro_NotZeroErr(fC, .001)) {
			Print_Point(sP1); Print_Point(sP2); Print_Point(sP3);
			printf("%f\n", fC); 
			printf("%f %f %f\n", sV.i, sV.j, sV.k);
			printf("%f %f %f\n", sPlane.sN.i, sPlane.sN.j, sPlane.sN.k);
			ERROR_Message("Plane_Point_Normal", "sV1 * sN != 0"); }
	sV.i = fXDiff[1]; sV.j = fYDiff[1]; sV.k = fZDiff[1];
	fC = Macro_DotProduct(sV, sPlane.sN);
	if (Macro_NotZeroErr(fC, .001)) {
			Print_Point(sP1); Print_Point(sP2); Print_Point(sP3);
			printf("%f\n", fC);
			printf("%f %f %f\n", sV.i, sV.j, sV.k);
			printf("%f %f %f\n", sPlane.sN.i, sPlane.sN.j, sPlane.sN.k);
			ERROR_Message("Plane_Point_Normal", "sV2 * sN != 0"); }
}
   /*************************************************************/

	return(sPlane.sN);
}

POINT_TYPE	Line_Plane_Intersection(PARA_LINE_TYPE sLine, PLANE_TYPE sPlane)
{
	VECTOR_TYPE		sV;
	float			t, fNumer, fDenom;
	POINT_TYPE		sP;

	fDenom = Macro_DotProduct(sLine.sV, sPlane.sN);
	if (Macro_Zero(fDenom))	 /* This provides an approximate point of intersection */ 
								/* since the two vectors are roughly parallel, or parallel */
	 	return(Nearest_Point_On_Plane(sPlane, ParaLine_Calc_Point(sLine.sV, sLine.sP, (float) 20000)));

	sV.i = sPlane.sP.fx - sLine.sP.fx;
	sV.j = sPlane.sP.fy - sLine.sP.fy;
	sV.k = sPlane.sP.fz - sLine.sP.fz;

	fNumer = Macro_DotProduct(sV, sPlane.sN);

	t = fNumer / fDenom;

		/* curtails t's too large.  Also makes sure that the intersection 
			is in front of the virtual transmitter */
	if (t > 20000 || t < 0)	t = (float) 20000;


	sP = Nearest_Point_On_Plane(sPlane, ParaLine_Calc_Point(sLine.sV, sLine.sP, t));

	return(sP);
}

UNIT_VECTOR Reflection_Vector(VECTOR_TYPE sV, PLANE_TYPE sPlane)
{
	VECTOR_TYPE 	sRefl;
	float			f;
	
	f = Macro_DotProduct(sV, sPlane.sN);
	sRefl.i = (float) -1 * (((float) 2 * (sPlane.sN.i * f)) - sV.i);
	sRefl.j = (float) -1 * (((float) 2 * (sPlane.sN.j * f)) - sV.j);
	sRefl.k = (float) -1 * (((float) 2 * (sPlane.sN.k * f)) - sV.k);

	return(Normalize_Vector(sRefl));
}

float Point_To_Plane_DistancePM(PLANE_TYPE sPlane, POINT_TYPE sP)
{
	VECTOR_TYPE		sV;
	float			fVal;
	
	sV.i = sP.fx - sPlane.sP.fx;
	sV.j = sP.fy - sPlane.sP.fy;
	sV.k = sP.fz - sPlane.sP.fz;

 		/* the plane's normal vector is assumed to be a unit vector */
	fVal = Macro_DotProduct(sV, sPlane.sN);

	return(fVal);	
}
float Point_To_Plane_Distance(PLANE_TYPE sPlane, POINT_TYPE sP)
{
	return((float) fabs(Point_To_Plane_DistancePM(sPlane, sP)));	
}
float Point_To_Plane_DistancePM2(UNIT_VECTOR sN, POINT_TYPE sP1, POINT_TYPE sP2)
{
	PLANE_TYPE	sPlane;

	sPlane.sN = sN;
	sPlane.sP = sP1;
	
	return(Point_To_Plane_DistancePM(sPlane, sP2));
}

short Point_On_Plane(PLANE_TYPE sPlane, POINT_TYPE sP)
{
	float	fVal=Point_To_Plane_DistancePM(sPlane, sP);
	
	if (Macro_ZeroErr(fVal, .001))
		return(1);
	else if (fVal > 0)
		return(0);
	else	
		return(-1);
}

POINT_TYPE Nearest_Point_On_Plane(PLANE_TYPE sPlane, POINT_TYPE sP)
{
	float			fVal;
	POINT_TYPE		sP2;
	
 		/* the plane's normal vector is assumed to be a unit vector */
	fVal = (float) -1 * Point_To_Plane_DistancePM(sPlane, sP);

	sP2.fx = sP.fx + fVal * sPlane.sN.i;
	sP2.fy = sP.fy + fVal * sPlane.sN.j;
	sP2.fz = sP.fz + fVal * sPlane.sN.k;

	/*if (Point_On_Plane(sPlane, sP2) != 1) 	{ printf("Hmmm"); Print_Point(sP); Print_Point(sP2); cc=getchar(); }*/

	return(sP2);	
}

short PointsEqual3D(POINT_TYPE sP1, POINT_TYPE sP2, float e)
{
	if (sP1.fx < sP2.fx + e && sP1.fx > sP2.fx - e &&
		sP1.fy < sP2.fy + e && sP1.fy > sP2.fy - e &&
		sP1.fz < sP2.fz + e && sP1.fz > sP2.fz - e)
		return(1);
	else
		return(0);
}
short VectorsEqual3D(VECTOR_TYPE sV1, VECTOR_TYPE sV2, float e)
{
	if (sV1.i < sV2.i + e && sV1.i > sV2.i - e &&
		sV1.j < sV2.j + e && sV1.j > sV2.j - e &&
		sV1.k < sV2.k + e && sV1.k > sV2.k - e)
		return(1);
	else
		return(0);
}

POINT_TYPE ComputeCentroid3D(POINT_TYPE asTriP[])
{
	POINT_TYPE	sC;
	float		fX = (float) 0, fY = (float) 0, fZ = (float) 0;
	short		i;

	for(i=0; i<3; i++) {
		fX = fX + asTriP[i].fx;
		fY = fY + asTriP[i].fy;
		fZ = fZ + asTriP[i].fz;
	}
	sC.fx = fX / 3;
	sC.fy = fY / 3;
	sC.fz = fZ / 3;

	return(sC);
}
	/* make general routines for this sometime */
POINT_TYPE Calc_Intersection_Of_Segments2D(POINT_TYPE sM1, POINT_TYPE sM2, 
												POINT_TYPE sN1, POINT_TYPE sN2, short* piFlag)
{	
	float			fSlope1, fB1, fSlope2, fB2, fVal;
	int				iS1=1, iS2=1;
 	POINT_TYPE		sInter;

	*piFlag = 0;

	fVal = sM1.fx - sM2.fx;
	if (Macro_NotZero(fVal)) {
		fSlope1 = (sM1.fy - sM2.fy) / fVal;
		fB1 = sM1.fy - fSlope1 * sM1.fx;
	}
	else
		iS1 = 0;
	fVal = sN1.fx - sN2.fx;
	if (Macro_NotZero(fVal)) {
		fSlope2 = (sN1.fy - sN2.fy) / fVal;
		fB2 = sN1.fy - fSlope2 * sN1.fx;
	}
	else
		iS2 = 0;
	if (iS1 == 1 && iS2 == 1 && Macro_Zero(fSlope1 - fSlope2))	{ *piFlag = 1; return(sM1); }
	if (iS1 == 1 && iS2 == 1) {
		sInter.fx = (fB2 - fB1) / (fSlope1 - fSlope2);
		sInter.fy = fSlope1 * sInter.fx + fB1;
	}
	else if (iS1 == 0 && iS2 == 1) {
		sInter.fx = sM1.fx;
		sInter.fy = fSlope2 * sInter.fx + fB2;
	}
	else if (iS1 == 1 && iS2 == 0) {
		sInter.fx = sN1.fx;
		sInter.fy = fSlope1 * sInter.fx + fB1;
	}
	else {	/* degenerate condition check */
		if (Between(sM1, sM2, sN1))
			sInter = sN1;
		else if (Between(sM1, sM2, sN2))
			sInter = sN2;
		else
			{ *piFlag = 1; printf("FLAGGGED!\n");}
	}

	return(sInter);
}

/***************************************************************************************
	Parse_Value:
		Finds the first value encountered and stores it in c.  c is assumed to
	be large enough to hold the value.
		
	- Assumes values are separated by spaces.

	- Returns a char* to the position starting right after the value ends
***************************************************************************************/
char* Parse_Value(char c[], char s[])
{
	int		k, l;

	k = 0;
	while (s[k] == ' ')		k++;

	l = 0;
	while (s[k] != ' ' && s[k] != '\0') 
		c[l++] = s[k++];
	c[l] = '\0';

	return(s + k);
}
int MyBinSearch(float fV, float fArr[], int iN, short iAscend)
{
	int		iLow=0, iUpp=iN, iMid=iN/2;
	int		iFound = -1;
	short	add;

	if (iAscend)	add=1; 	else add=-1;

	while (iFound < 0)
	{
		if (fV >= fArr[iMid] || iMid == iN || iMid == 0)
			if (iMid+add > iN || iMid+add < 0)
				if (iAscend) 	iFound = iN; 	else 	iFound = 0;
			else if (fV < fArr[iMid+add])
				if (fV - fArr[iMid] < fArr[iMid+add] - fV)  
					iFound = iMid;  
				else 
					iFound = iMid+add;
			else
				if (iAscend) 
					{ iLow = iMid; iMid = (iUpp + iLow + 1)/2; }
				else
					{ iUpp = iMid; iMid = (iUpp + iLow)/2; }
		else
			if (iAscend) 
				{ iUpp = iMid; iMid = (iUpp + iLow)/2; }
			else
				{ iLow = iMid; iMid = (iUpp + iLow + 1)/2; }
	}

	return(iFound);
}
